from ark_algebra_py.ark_algebra_py import Scalar

# Initialisation - The default initialiser for a scalar is an u128 integer
scalar = Scalar(12345)

# Equality -- We override eq and neq operators
assert scalar == scalar
assert Scalar(1234) != Scalar(4567)

# Scalar Addition/subtraction/Negation -- We override the add/sub/neg operators
a = Scalar(3)
b = Scalar(4)
c = Scalar(5)
assert a.square() + b.square() == c.square()
assert a * a + b * b == c * c

neg_a = -a
assert a + neg_a == Scalar(0)
assert (a + neg_a).is_zero()

# Serialisation
compressed_bytes = scalar.to_le_bytes()
deserialised_scalar = Scalar.from_le_bytes(compressed_bytes)
assert scalar == deserialised_scalar


import ark_algebra_py
from ark_algebra_py import ark_algebra_py

__doc__ = ark_algebra_py.ark_algebra_py.__doc__
if hasattr(ark_algebra_py.ark_algebra_py, "__all__"):
    __all__ = ark_algebra_py.ark_algebra_py.__all__


from ark_algebra_py.ark_algebra_py import Polynomial, Scalar, Domain


# Initilisation -- This is the generator point
a = Polynomial([Scalar(100), Scalar(10), Scalar(1)])
b = Polynomial([Scalar(10), Scalar(5), Scalar(1)])

print(a + b)
print(a - b)
print(a * b)
print(a * b)

assert a.degree() == 2
assert b.degree() == 2
assert (a + b).degree() == 2

a_at_2 = a.evaluate(Scalar(2))
b_at_2 = b.evaluate(Scalar(2))

assert a_at_2 == Scalar(124)
assert b_at_2 == Scalar(24)
assert a_at_2 + b_at_2 == (a + b).evaluate(Scalar(2))
assert a_at_2 * b_at_2 == (a * b).evaluate(Scalar(2))

domain = Domain(8)
a = domain.interpolate([Scalar(1), Scalar(2), Scalar(3)])

assert(a.evaluate(domain.element(0)) == Scalar(1))
assert(a.evaluate(domain.element(1)) == Scalar(2))
assert(a.evaluate(domain.element(2)) == Scalar(3))



from py_ecc.bls import G2ProofOfPossession as py_ecc_bls
from py_ecc.optimized_bls12_381 import (  # noqa: F401
    G1,
    G2,
    Z1,
    add as py_ecc_add,
    multiply as py_ecc_mul,
    neg as py_ecc_neg,
    pairing as py_ecc_pairing,
    final_exponentiate as py_ecc_final_exponentiate,
    FQ12 as py_ecc_GT
)
from py_ecc.bls.g2_primitives import (  # noqa: F401
    G1_to_pubkey as py_ecc_G1_to_bytes48,
    pubkey_to_G1 as py_ecc_bytes48_to_G1,
    G2_to_signature as py_ecc_G2_to_bytes96,
    signature_to_G2 as py_ecc_bytes96_to_G2,
)


def pairing_check(values):
    p_q_1, p_q_2 = values
    final_exponentiation = py_ecc_final_exponentiate(
        py_ecc_pairing(p_q_1[1], p_q_1[0], final_exponentiate=False)
        * py_ecc_pairing(p_q_2[1], p_q_2[0], final_exponentiate=False)
    )
    return final_exponentiation == py_ecc_GT.one()


# Performs point addition of `lhs` and `rhs`
# The points can either be in G1 or G2
def add(lhs, rhs):
    return py_ecc_add(lhs, rhs)


# Performs Scalar multiplication between
# `point` and `scalar`
# `point` can either be in G1 or G2
def multiply(point, scalar):
    return py_ecc_mul(point, scalar)


# Returns the point negation of `point`
# `point` can either be in G1 or G2
def neg(point):
    return py_ecc_neg(point)


# Serializes a point in G1
# Returns a bytearray of size 48 as
# we use the compressed format
def G1_to_bytes48(point):
    return py_ecc_G1_to_bytes48(point)


# Serializes a point in G2
# Returns a bytearray of size 96 as
# we use the compressed format
def G2_to_bytes96(point):
    return py_ecc_G2_to_bytes96(point)


# Deserializes a purported compressed serialized
# point in G1
# - No subgroup checks are performed
# - If the bytearray is not a valid serialization
# of a point in G1, then this method will raise
# an exception
def bytes48_to_G1(bytes48):
    return py_ecc_bytes48_to_G1(bytes48)


# Deserializes a purported compressed serialized
# point in G2
# - No subgroup checks are performed
# - If the bytearray is not a valid serialization
# of a point in G2, then this method will raise
# an exception
def bytes96_to_G2(bytes96):
    return py_ecc_bytes96_to_G2(bytes96)






from ark_algebra_py.ark_algebra_py import G1, G2, Scalar

# G1Point and G2Point have the same methods implemented on them
# For brevity, I will only show one method using G1Point and G2Point 
# The rest of the code will just use G1Point

# Point initialization -- This will be initialized to the g1 generator 
g1_generator = G1()
g2_generator = G2()

# Identity element 
identity = G1.identity()

# Equality -- We override eq and neq operators
assert g1_generator == g1_generator
assert g1_generator != identity


# Printing an element -- We override __str__ so when we print
# an element it prints in hex
print("identity: ",identity)
print("g1 generator: ", g1_generator)
print("g2 generator: ", g2_generator)

# Point Addition/subtraction/Negation -- We override the add/sub/neg operators
gen = G1()
double_gen = gen + gen
double_gen2 = gen.double()
assert double_gen == double_gen2
assert (double_gen2 - gen) == gen
neg_gen = -gen
assert neg_gen + gen == identity

# Scalar multiplication
scalar = Scalar(4)
four_gen = gen * scalar
four_gen_2 = scalar * gen 
assert four_gen == gen + gen + gen + gen
assert four_gen == four_gen_2

# Serialisation
# 
# serialising to/from a g1 point
# We don't expose the uncompressed form 
# because it seems like its not needed
compressed_bytes = gen.to_compressed_bytes()
deserialised_point = G1.from_compressed_bytes(compressed_bytes)
# If the bytes being received are trusted, we can avoid
# doing subgroup checks
deserialised_point_unchecked = G1.from_compressed_bytes_unchecked(compressed_bytes)
assert deserialised_point == deserialised_point_unchecked
assert deserialised_point == gen



from py_arkworks_bls12381 import G1Point, G2Point, Scalar, GT
from typing import Sequence, Tuple

# Copies the interface for py-ecc

# This is the identity point or point at infinity
Z1 = G1Point.identity()
# Generator point for G1
G1 = G1Point()
# Generator point for G2
G2 = G2Point()

def add(lhs : G1Point | G2Point, rhs : G1Point | G1Point) -> G1Point | G2Point:
    return lhs + rhs

def neg(point : G1Point | G2Point) -> G1Point | G2Point:
    return -point

def multiply(point : G1Point | G2Point, integer : int) -> G1Point | G2Point:
    int_as_bytes = integer.to_bytes(32, 'little')
    scalar = Scalar.from_le_bytes(int_as_bytes)
    return point * scalar

def pairing_check(values : Sequence[Tuple[G1Point, G2Point]]) -> bool:
    p_q_1, p_q_2 = values
    g1s = [p_q_1[0], p_q_2[0]]
    g2s = [p_q_1[1], p_q_2[1]]
    return GT.multi_pairing(g1s, g2s) == GT.one()

def G1_to_bytes48(point : G1Point) -> bytes:
    return point.to_compressed_bytes()

def G2_to_bytes96(point : G2Point) -> bytes:
    return point.to_compressed_bytes()

def bytes96_to_G2(bytes96) -> G1Point:
    # This will not do subgroup checks
    # If you want that then remove the `_unchecked` prefix
    return G2Point.from_compressed_bytes_unchecked(bytes96)

def bytes48_to_G1(bytes48) -> G1Point:
    # This will not do subgroup checks
    # If you want that then remove the `_unchecked` prefix
    return G1Point.from_compressed_bytes_unchecked(bytes48)


# Small test code to make sure it compiles and show
# you how to call the methods
bytes48 = G1_to_bytes48(G1Point())
g1_point = bytes48_to_G1(bytes48)

bytes96 = G2_to_bytes96(G2Point())
g2_point = bytes96_to_G2(bytes96)

neg_g1 = neg(g1_point)

expected_identity = add(neg_g1, g1_point)
assert expected_identity == G1Point.identity()

four_gen = multiply(G2, 4)
expected_four_gen = G2 + G2 + G2 + G2
assert expected_four_gen == four_gen

a = Scalar(20)
b = Scalar(30)
ok = pairing_check([
    [G1 * a, G2 * b],
    [G1 * -a, G2 * b]
])
assert ok


from .arkworks import G1 as ark_G1, G2 as ark_G2, pairing_check as ark_pairing
from .py_ecc_4844 import G1 as py_G1, G2 as py_G2,pairing_check as py_pairing

from timeit import timeit

def timeit_dic(module_name):
   return {
   0: {
      "method_name": "pairingsGen",
      "setup": "from {} import G1, G2, pairing_check".format(module_name),
      "statement": "pairing_check([[G1, G2], [G1, G2]])",
      "iterations": 10,
   },
   1: {
      "method_name": "pairings",
      "setup": "from {} import G1, G2, pairing_check, multiply; a = multiply(G1, 123456789);b = multiply(G2, 123456789)".format(module_name),
      "statement": "pairing_check([[G1, G2], [a, b]])",
      "iterations": 10,
   },
   2: {
      "method_name": "multiplyG1",
      "setup": "from {} import G1, multiply".format(module_name),
      "statement": "multiply(G1, 12345678910)",
      "iterations": 100,
   },
   4: {
      "method_name": "multiplyG2",
      "setup": "from {} import G2, multiply".format(module_name),
      "statement": "multiply(G2, 12345678910)",
      "iterations": 100,
   },
   5: {
      "method_name": "addG1",
      "setup": "from {} import G1, G2, add".format(module_name),
      "statement": "add(G1, G1)",
      "iterations": 100,
   },
   6: {
      "method_name": "addG2",
      "setup": "from {} import G2, add".format(module_name),
      "statement": "add(G2, G2)",
      "iterations": 100,
   },
   7: {
      "method_name": "G1_to_bytes48",
      "setup": "from {} import G1,G1_to_bytes48".format(module_name),
      "statement": "G1_to_bytes48(G1)",
      "iterations": 100,
   },
   8: {
      "method_name": "G2_to_bytes96",
      "setup": "from {} import G2,G2_to_bytes96".format(module_name),
      "statement": "G2_to_bytes96(G2)",
      "iterations": 100,
   },
   }

def run_benchmark(module_name):
   # Create timeit dictionary to specify
   # all of the benchmarks we want to run
   # py_ecc and arkworks have the same api
   # so we can template the dictionary and just 
   # pass the module name
   benches = timeit_dic(module_name)
   
   print(f"\n{module_name} benchmarks\n") 
   
   # Run benchmarks
   for id, library in benches.items():
      result = timeit(
         stmt=library["statement"],
         setup=library["setup"],
         number=library["iterations"],
         )
      # multiply by 1000 to turn seconds into milliseconds
      print(f"{library['method_name']}", (result / library["iterations"])*1000, "ms")

run_benchmark("arkworks")
run_benchmark("py_ecc_4844")


from ark_algebra_py.ark_algebra_py import G1, G2, PairingOutput, Scalar, Pairing


# Initilisation -- This is the generator point
gt_gen = PairingOutput()

# Zero/One
zero = PairingOutput.one()

# Computing a pairing using pairing and multi_pairing
# multi_pairing does multiple pairings and adds them together with only one final_exp
assert gt_gen == Pairing.pairing(G1(), G2()) 
g1s = [G1()]
g2s = [G2()]
assert gt_gen == Pairing.multi_pairing(g1s, g2s)

# Bilinearity
a = Scalar(1234)
b = Scalar(4566)
c = a * b


g = G1() * a
h = G2() * b

p = Pairing.pairing(g, h)

c_g1 = G1() *c
c_g2 = G2() *c

assert p == Pairing.pairing(c_g1, G2())
assert p == Pairing.pairing(G1(), c_g2)
